// SPDX-License-Identifier: LGPL-3.0-or-later
// See Notices.txt for copyright information

//! Fayalite Simulation

use crate::{
    bundle::{BundleField, BundleType},
    expr::{
        ops,
        target::{
            GetTarget, Target, TargetBase, TargetPathArrayElement, TargetPathBundleField,
            TargetPathElement,
        },
        ExprEnum, Flow,
    },
    intern::{Intern, Interned, Memoize},
    module::{
        AnnotatedModuleIO, Block, ModuleBody, NormalModuleBody, Stmt, StmtConnect, StmtDeclaration,
        StmtFormal, StmtIf, StmtInstance, StmtMatch, StmtReg, StmtWire,
    },
    prelude::*,
    sim::interpreter::{
        Insn, InsnField, InsnFieldKind, InsnFieldType, Insns, InsnsBuilding, InsnsBuildingDone,
        SlotDebugData, StatePartArrayIndex, StatePartArrayIndexed, StatePartIndex,
        StatePartIndexMap, StatePartIndexRange, StatePartKind, StatePartKindBigSlots,
        StatePartKindSmallSlots, StatePartLayout, StatePartsValue, TypeArrayIndex,
        TypeArrayIndexes, TypeIndex, TypeIndexRange, TypeLayout, TypeParts,
    },
    ty::StaticType,
};
use hashbrown::HashMap;
use num_bigint::BigInt;
use std::{collections::BTreeSet, fmt};

mod interpreter;

#[derive(Debug, PartialEq, Eq, Hash, Clone, Copy)]
enum CondBody {
    IfTrue {
        cond: CompiledExpr<Bool>,
    },
    IfFalse {
        cond: CompiledExpr<Bool>,
    },
    MatchArm {
        enum_expr: CompiledExpr<Enum>,
        variant_index: usize,
    },
}

#[derive(Debug, PartialEq, Eq, Hash, Clone, Copy)]
struct Cond {
    body: CondBody,
    source_location: SourceLocation,
}

#[derive(PartialEq, Eq, Hash, Clone, Copy)]
enum InstantiatedModule {
    Base(Interned<Module<Bundle>>),
    Child {
        parent: Interned<InstantiatedModule>,
        instance: Interned<Instance<Bundle>>,
    },
}

impl InstantiatedModule {
    fn leaf_module(self) -> Interned<Module<Bundle>> {
        match self {
            InstantiatedModule::Base(base) => base,
            InstantiatedModule::Child { instance, .. } => instance.instantiated(),
        }
    }
    fn write_path(self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
        match self {
            InstantiatedModule::Base(base) => fmt::Debug::fmt(&base.name_id(), f),
            InstantiatedModule::Child { parent, instance } => {
                parent.write_path(f)?;
                write!(f, ".{}", instance.name_id())
            }
        }
    }
}

impl fmt::Debug for InstantiatedModule {
    fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
        write!(f, "InstantiatedModule(")?;
        self.write_path(f)?;
        write!(f, ": {})", self.leaf_module().name_id())
    }
}

#[derive(Debug, PartialEq, Eq, Hash, Clone, Copy)]
struct TargetInInstantiatedModule {
    instantiated_module: InstantiatedModule,
    target: Target,
}

#[derive(Debug, PartialEq, Eq, Hash, Clone, Copy)]
struct CompiledBundleField {
    offset: TypeIndex,
    ty: CompiledTypeLayout<CanonicalType>,
}

#[derive(Debug, PartialEq, Eq, Hash, Clone, Copy)]
enum CompiledTypeLayoutBody {
    Scalar,
    Array {
        /// debug names are ignored, use parent's layout instead
        element: Interned<CompiledTypeLayout<CanonicalType>>,
    },
    Bundle {
        /// debug names are ignored, use parent's layout instead
        fields: Interned<[CompiledBundleField]>,
    },
}

#[derive(Debug, PartialEq, Eq, Hash, Clone, Copy)]
struct CompiledTypeLayout<T: Type> {
    ty: T,
    layout: TypeLayout<InsnsBuildingDone>,
    body: CompiledTypeLayoutBody,
}

impl<T: Type> CompiledTypeLayout<T> {
    fn with_prefixed_debug_names(self, prefix: &str) -> Self {
        let Self { ty, layout, body } = self;
        Self {
            ty,
            layout: layout.with_prefixed_debug_names(prefix),
            body,
        }
    }
    fn with_anonymized_debug_info(self) -> Self {
        let Self { ty, layout, body } = self;
        Self {
            ty,
            layout: layout.with_anonymized_debug_info(),
            body,
        }
    }
    fn get(ty: T) -> Self {
        #[derive(Clone, Copy, PartialEq, Eq, Hash, Debug)]
        struct MyMemoize;
        impl Memoize for MyMemoize {
            type Input = CanonicalType;
            type InputOwned = CanonicalType;
            type Output = CompiledTypeLayout<CanonicalType>;

            fn inner(self, input: &Self::Input) -> Self::Output {
                match input {
                    CanonicalType::UInt(_)
                    | CanonicalType::SInt(_)
                    | CanonicalType::Bool(_)
                    | CanonicalType::Enum(_)
                    | CanonicalType::AsyncReset(_)
                    | CanonicalType::SyncReset(_)
                    | CanonicalType::Reset(_)
                    | CanonicalType::Clock(_) => {
                        let mut layout = TypeLayout::empty();
                        let debug_data = SlotDebugData {
                            name: Interned::default(),
                            ty: *input,
                        };
                        layout.big_slots = StatePartLayout::scalar(debug_data);
                        CompiledTypeLayout {
                            ty: *input,
                            layout: layout.into(),
                            body: CompiledTypeLayoutBody::Scalar,
                        }
                    }
                    CanonicalType::Array(array) => {
                        let mut layout = TypeLayout::empty();
                        let element = CompiledTypeLayout::get(array.element()).intern_sized();
                        for index in 0..array.len() {
                            layout.allocate(
                                &element
                                    .layout
                                    .with_prefixed_debug_names(&format!("[{index}]")),
                            );
                        }
                        CompiledTypeLayout {
                            ty: *input,
                            layout: layout.into(),
                            body: CompiledTypeLayoutBody::Array { element },
                        }
                    }
                    CanonicalType::Bundle(bundle) => {
                        let mut layout = TypeLayout::empty();
                        let fields = bundle
                            .fields()
                            .iter()
                            .map(
                                |BundleField {
                                     name,
                                     flipped: _,
                                     ty,
                                 }| {
                                    let ty = CompiledTypeLayout::get(*ty);
                                    let offset = layout
                                        .allocate(
                                            &ty.layout
                                                .with_prefixed_debug_names(&format!(".{name}")),
                                        )
                                        .start();
                                    CompiledBundleField { offset, ty }
                                },
                            )
                            .collect();
                        CompiledTypeLayout {
                            ty: *input,
                            layout: layout.into(),
                            body: CompiledTypeLayoutBody::Bundle { fields },
                        }
                    }
                }
            }
        }
        let CompiledTypeLayout {
            ty: _,
            layout,
            body,
        } = MyMemoize.get_owned(ty.canonical());
        Self { ty, layout, body }
    }
}

#[derive(Debug, PartialEq, Eq, Hash, Clone, Copy)]
struct CompiledValue<T: Type> {
    layout: CompiledTypeLayout<T>,
    range: TypeIndexRange,
    write: Option<(CompiledTypeLayout<T>, TypeIndexRange)>,
}

impl<T: Type> CompiledValue<T> {
    fn write(self) -> (CompiledTypeLayout<T>, TypeIndexRange) {
        self.write.unwrap_or((self.layout, self.range))
    }
    fn map<U: Type>(
        self,
        mut f: impl FnMut(
            CompiledTypeLayout<T>,
            TypeIndexRange,
        ) -> (CompiledTypeLayout<U>, TypeIndexRange),
    ) -> CompiledValue<U> {
        let (layout, range) = f(self.layout, self.range);
        CompiledValue {
            layout,
            range,
            write: self.write.map(|(layout, range)| f(layout, range)),
        }
    }
    fn map_ty<U: Type>(self, mut f: impl FnMut(T) -> U) -> CompiledValue<U> {
        self.map(|CompiledTypeLayout { ty, layout, body }, range| {
            (
                CompiledTypeLayout {
                    ty: f(ty),
                    layout,
                    body,
                },
                range,
            )
        })
    }
}

impl CompiledValue<Enum> {
    fn add_discriminant_to_set(self, offset: TypeIndex, inputs: &mut SlotSet) {
        inputs.extend([self.range.offset(offset)]);
    }
}

impl CompiledValue<Bundle> {
    fn field_by_index(self, field_index: usize) -> CompiledValue<CanonicalType> {
        self.map(|layout, range| {
            let CompiledTypeLayout {
                ty: _,
                layout: _,
                body: CompiledTypeLayoutBody::Bundle { fields },
            } = layout
            else {
                unreachable!();
            };
            (
                fields[field_index].ty,
                range.slice(TypeIndexRange::new(
                    fields[field_index].offset,
                    fields[field_index].ty.layout.len(),
                )),
            )
        })
    }
    fn field_by_name(self, name: Interned<str>) -> CompiledValue<CanonicalType> {
        self.field_by_index(self.layout.ty.name_indexes()[&name])
    }
}

impl CompiledValue<Array> {
    fn element(self, index: usize) -> CompiledValue<CanonicalType> {
        self.map(|layout, range| {
            let CompiledTypeLayoutBody::Array { element } = layout.body else {
                unreachable!();
            };
            (*element, range.index_array(element.layout.len(), index))
        })
    }
    fn element_dyn(
        self,
        index_slot: StatePartIndex<StatePartKindSmallSlots>,
    ) -> CompiledExpr<CanonicalType> {
        CompiledExpr::from(self).element_dyn(index_slot)
    }
}

#[derive(Debug, PartialEq, Eq, Hash, Copy, Clone)]
struct CompiledExpr<T: Type> {
    static_part: CompiledValue<T>,
    indexes: TypeArrayIndexes,
}

impl<T: Type> From<CompiledValue<T>> for CompiledExpr<T> {
    fn from(static_part: CompiledValue<T>) -> Self {
        Self {
            static_part,
            indexes: TypeArrayIndexes::default(),
        }
    }
}

impl<T: Type> CompiledExpr<T> {
    fn map_ty<U: Type>(self, f: impl FnMut(T) -> U) -> CompiledExpr<U> {
        let Self {
            static_part,
            indexes,
        } = self;
        CompiledExpr {
            static_part: static_part.map_ty(f),
            indexes,
        }
    }
    fn add_target_without_indexes_to_set(self, inputs: &mut SlotSet) {
        let Self {
            static_part,
            indexes,
        } = self;
        indexes.as_ref().for_each_offset(|offset| {
            inputs.extend([static_part.range.offset(offset)]);
        });
    }
    fn add_target_and_indexes_to_set(self, inputs: &mut SlotSet) {
        let Self {
            static_part: _,
            indexes,
        } = self;
        self.add_target_without_indexes_to_set(inputs);
        inputs.extend(indexes.as_ref().iter());
    }
}

impl CompiledExpr<Enum> {
    fn add_discriminant_and_indexes_to_set(self, inputs: &mut SlotSet) {
        let Self {
            static_part,
            indexes,
        } = self;
        indexes.as_ref().for_each_offset(|offset| {
            static_part.add_discriminant_to_set(offset, inputs);
        });
        inputs.extend(indexes.as_ref().iter());
    }
}

impl CompiledExpr<Bundle> {
    fn field_by_index(self, field_index: usize) -> CompiledExpr<CanonicalType> {
        CompiledExpr {
            static_part: self.static_part.field_by_index(field_index),
            indexes: self.indexes,
        }
    }
    fn field_by_name(self, name: Interned<str>) -> CompiledExpr<CanonicalType> {
        CompiledExpr {
            static_part: self.static_part.field_by_name(name),
            indexes: self.indexes,
        }
    }
}

impl CompiledExpr<Array> {
    fn element(self, index: usize) -> CompiledExpr<CanonicalType> {
        CompiledExpr {
            static_part: self.static_part.element(index),
            indexes: self.indexes,
        }
    }
    fn element_dyn(
        self,
        index_slot: StatePartIndex<StatePartKindSmallSlots>,
    ) -> CompiledExpr<CanonicalType> {
        let CompiledTypeLayoutBody::Array { element } = self.static_part.layout.body else {
            unreachable!();
        };
        let stride = element.layout.len();
        let indexes = self.indexes.join(TypeArrayIndex::from_parts(
            index_slot,
            self.static_part.layout.ty.len(),
            stride,
        ));
        CompiledExpr {
            static_part: self.static_part.map(|layout, range| {
                let CompiledTypeLayoutBody::Array { element } = layout.body else {
                    unreachable!();
                };
                (*element, range.index_array(stride, 0))
            }),
            indexes,
        }
    }
}

#[derive(Clone, Debug, Default, PartialEq, Eq, Hash)]
struct SlotSet(TypeParts<Self>);

impl SlotSet {
    fn is_empty(&self) -> bool {
        let Self(TypeParts {
            small_slots,
            big_slots,
        }) = self;
        small_slots.is_empty() && big_slots.is_empty()
    }
}

impl StatePartsValue for SlotSet {
    type Value<K: StatePartKind> = BTreeSet<StatePartIndex<K>>;
}

impl Extend<StatePartIndex<StatePartKindSmallSlots>> for SlotSet {
    fn extend<T: IntoIterator<Item = StatePartIndex<StatePartKindSmallSlots>>>(&mut self, iter: T) {
        self.0.small_slots.extend(iter);
    }
}

impl Extend<StatePartIndex<StatePartKindBigSlots>> for SlotSet {
    fn extend<T: IntoIterator<Item = StatePartIndex<StatePartKindBigSlots>>>(&mut self, iter: T) {
        self.0.big_slots.extend(iter);
    }
}

impl<K: StatePartKind> Extend<StatePartIndexRange<K>> for SlotSet
where
    Self: Extend<StatePartIndex<K>>,
{
    fn extend<T: IntoIterator<Item = StatePartIndexRange<K>>>(&mut self, iter: T) {
        self.extend(iter.into_iter().flat_map(|v| v.iter()));
    }
}

impl Extend<TypeIndexRange> for SlotSet {
    fn extend<T: IntoIterator<Item = TypeIndexRange>>(&mut self, iter: T) {
        iter.into_iter().for_each(
            |TypeIndexRange {
                 small_slots,
                 big_slots,
             }| {
                self.extend(small_slots.iter());
                self.extend(big_slots.iter());
            },
        )
    }
}

impl Extend<TypeArrayIndex> for SlotSet {
    fn extend<T: IntoIterator<Item = TypeArrayIndex>>(&mut self, iter: T) {
        iter.into_iter().for_each(
            |TypeArrayIndex {
                 small_slots,
                 big_slots,
             }| {
                self.extend([small_slots]);
                self.extend([big_slots]);
            },
        )
    }
}

impl<K: StatePartKind> Extend<StatePartArrayIndex<K>> for SlotSet {
    fn extend<T: IntoIterator<Item = StatePartArrayIndex<K>>>(&mut self, iter: T) {
        self.extend(iter.into_iter().map(|v| v.index));
    }
}

impl Extend<CondBody> for SlotSet {
    fn extend<T: IntoIterator<Item = CondBody>>(&mut self, iter: T) {
        iter.into_iter().for_each(|cond_body| match cond_body {
            CondBody::IfTrue { cond } | CondBody::IfFalse { cond } => {
                cond.add_target_and_indexes_to_set(self);
            }
            CondBody::MatchArm {
                enum_expr,
                variant_index,
            } => enum_expr.add_discriminant_and_indexes_to_set(self),
        })
    }
}

impl Extend<Cond> for SlotSet {
    fn extend<T: IntoIterator<Item = Cond>>(&mut self, iter: T) {
        self.extend(iter.into_iter().map(|v| v.body))
    }
}

#[derive(Debug)]
struct Assignment {
    inputs: SlotSet,
    conditions: Interned<[Cond]>,
    insns: Vec<Insn>,
    source_location: SourceLocation,
}

#[derive(Debug, Default)]
struct StatePartAssignments<K: StatePartKind> {
    written_slot_to_assignment_indexes_map: StatePartIndexMap<K, Vec<usize>>,
}

#[derive(Debug, Default)]
struct SlotAssignments {
    assignments: Vec<Assignment>,
    parts: TypeParts<Self>,
}

impl SlotAssignments {
    fn for_assignment(&mut self, assignment_index: usize) -> SlotAssignmentsForAssignment<'_> {
        SlotAssignmentsForAssignment {
            parts: &mut self.parts,
            assignment_index,
        }
    }
}

impl StatePartsValue for SlotAssignments {
    type Value<K: StatePartKind> = StatePartAssignments<K>;
}

struct SlotAssignmentsForAssignment<'a> {
    parts: &'a mut TypeParts<SlotAssignments>,
    assignment_index: usize,
}

impl Extend<StatePartIndex<StatePartKindSmallSlots>> for SlotAssignmentsForAssignment<'_> {
    fn extend<T: IntoIterator<Item = StatePartIndex<StatePartKindSmallSlots>>>(&mut self, iter: T) {
        iter.into_iter().for_each(|slot| {
            self.parts
                .small_slots
                .written_slot_to_assignment_indexes_map
                .entry(slot)
                .or_insert_with(Vec::new)
                .push(self.assignment_index)
        });
    }
}

impl Extend<StatePartIndex<StatePartKindBigSlots>> for SlotAssignmentsForAssignment<'_> {
    fn extend<T: IntoIterator<Item = StatePartIndex<StatePartKindBigSlots>>>(&mut self, iter: T) {
        iter.into_iter().for_each(|slot| {
            self.parts
                .big_slots
                .written_slot_to_assignment_indexes_map
                .entry(slot)
                .or_insert_with(Vec::new)
                .push(self.assignment_index)
        });
    }
}

#[derive(Debug)]
pub struct Compiler {
    insns: Insns<InsnsBuilding>,
    base_module: Interned<Module<Bundle>>,
    modules: HashMap<InstantiatedModule, CompiledModule>,
    compiled_values: HashMap<TargetInInstantiatedModule, CompiledValue<CanonicalType>>,
    compiled_exprs: HashMap<Expr<CanonicalType>, CompiledExpr<CanonicalType>>,
    compiled_exprs_to_values: HashMap<CompiledExpr<CanonicalType>, CompiledValue<CanonicalType>>,
    decl_conditions: HashMap<TargetInInstantiatedModule, Interned<[Cond]>>,
    slots_assignments: SlotAssignments,
}

impl Compiler {
    pub fn new(base_module: Interned<Module<Bundle>>) -> Self {
        Self {
            insns: Insns::new(),
            base_module,
            modules: HashMap::new(),
            compiled_values: HashMap::new(),
            compiled_exprs: HashMap::new(),
            compiled_exprs_to_values: HashMap::new(),
            decl_conditions: HashMap::new(),
            slots_assignments: SlotAssignments::default(),
        }
    }
    fn compile_value(
        &mut self,
        target: TargetInInstantiatedModule,
    ) -> CompiledValue<CanonicalType> {
        if let Some(&retval) = self.compiled_values.get(&target) {
            return retval;
        }
        let retval = match target.target {
            Target::Base(base) => {
                let unprefixed_layout = CompiledTypeLayout::get(base.canonical_ty());
                let layout = unprefixed_layout.with_prefixed_debug_names(&format!(
                    "{:?}.{:?}",
                    target.instantiated_module,
                    base.target_name()
                ));
                let range = self.insns.allocate_variable(&layout.layout);
                let write = match *base {
                    TargetBase::ModuleIO(_)
                    | TargetBase::MemPort(_)
                    | TargetBase::Wire(_)
                    | TargetBase::Instance(_) => None,
                    TargetBase::Reg(_) => {
                        let write_layout = unprefixed_layout.with_prefixed_debug_names(&format!(
                            "{:?}.{:?}$next",
                            target.instantiated_module,
                            base.target_name()
                        ));
                        Some((
                            write_layout,
                            self.insns.allocate_variable(&write_layout.layout),
                        ))
                    }
                };
                CompiledValue {
                    range,
                    layout,
                    write,
                }
            }
            Target::Child(target_child) => {
                let parent = self.compile_value(TargetInInstantiatedModule {
                    instantiated_module: target.instantiated_module,
                    target: *target_child.parent(),
                });
                match *target_child.path_element() {
                    TargetPathElement::BundleField(TargetPathBundleField { name }) => {
                        parent.map_ty(Bundle::from_canonical).field_by_name(name)
                    }
                    TargetPathElement::ArrayElement(TargetPathArrayElement { index }) => {
                        parent.map_ty(Array::from_canonical).element(index)
                    }
                    TargetPathElement::DynArrayElement(_) => unreachable!(),
                }
            }
        };
        self.compiled_values.insert(target, retval);
        retval
    }
    fn compiled_expr_to_value(
        &mut self,
        expr: CompiledExpr<CanonicalType>,
        source_location: SourceLocation,
    ) -> CompiledValue<CanonicalType> {
        if let Some(&retval) = self.compiled_exprs_to_values.get(&expr) {
            return retval;
        }
        assert!(
            expr.static_part.layout.ty.is_passive(),
            "invalid expression passed to compiled_expr_to_value -- type must be passive",
        );
        let CompiledExpr {
            static_part,
            indexes,
        } = expr;
        let layout = static_part.layout.with_anonymized_debug_info();
        let retval = CompiledValue {
            layout,
            range: self.insns.allocate_variable(&layout.layout),
            write: None,
        };
        let TypeIndexRange {
            small_slots,
            big_slots,
        } = retval.range;
        self.add_assignment(
            Interned::default(),
            small_slots
                .iter()
                .zip(static_part.range.small_slots.iter())
                .map(|(dest, base)| {
                    if indexes.small_slots.is_empty() {
                        Insn::CopySmall { dest, src: base }
                    } else {
                        Insn::ReadSmallIndexed {
                            dest,
                            src: StatePartArrayIndexed {
                                base,
                                indexes: indexes.small_slots,
                            },
                        }
                    }
                })
                .chain(
                    big_slots
                        .iter()
                        .zip(static_part.range.big_slots.iter())
                        .map(|(dest, base)| {
                            if indexes.big_slots.is_empty() {
                                Insn::Copy { dest, src: base }
                            } else {
                                Insn::ReadIndexed {
                                    dest,
                                    src: StatePartArrayIndexed {
                                        base,
                                        indexes: indexes.big_slots,
                                    },
                                }
                            }
                        }),
                ),
            source_location,
        );
        self.compiled_exprs_to_values.insert(expr, retval);
        retval
    }
    fn add_assignment(
        &mut self,
        conditions: Interned<[Cond]>,
        insns: impl IntoIterator<Item = Insn>,
        source_location: SourceLocation,
    ) {
        let insns = Vec::from_iter(insns);
        let assignment_index = self.slots_assignments.assignments.len();
        let mut inputs = SlotSet::default();
        for insn in &insns {
            for InsnField { ty, kind } in insn.fields() {
                match (kind, ty) {
                    (InsnFieldKind::Input, InsnFieldType::SmallSlot(&slot)) => {
                        inputs.extend([slot]);
                    }
                    (InsnFieldKind::Input, InsnFieldType::BigSlot(&slot)) => {
                        inputs.extend([slot]);
                    }
                    (
                        InsnFieldKind::Input,
                        InsnFieldType::SmallSlotArrayIndexed(&array_indexed),
                    ) => {
                        array_indexed.for_each_target(|slot| inputs.extend([slot]));
                        inputs.extend(array_indexed.indexes);
                    }
                    (InsnFieldKind::Input, InsnFieldType::BigSlotArrayIndexed(&array_indexed)) => {
                        array_indexed.for_each_target(|slot| inputs.extend([slot]));
                        inputs.extend(array_indexed.indexes);
                    }
                    (InsnFieldKind::Output, InsnFieldType::SmallSlot(&slot)) => self
                        .slots_assignments
                        .for_assignment(assignment_index)
                        .extend([slot]),
                    (InsnFieldKind::Output, InsnFieldType::BigSlot(&slot)) => self
                        .slots_assignments
                        .for_assignment(assignment_index)
                        .extend([slot]),
                    (
                        InsnFieldKind::Output,
                        InsnFieldType::SmallSlotArrayIndexed(&array_indexed),
                    ) => {
                        array_indexed.for_each_target(|slot| {
                            self.slots_assignments
                                .for_assignment(assignment_index)
                                .extend([slot])
                        });
                        inputs.extend(array_indexed.indexes);
                    }
                    (InsnFieldKind::Output, InsnFieldType::BigSlotArrayIndexed(&array_indexed)) => {
                        array_indexed.for_each_target(|slot| {
                            self.slots_assignments
                                .for_assignment(assignment_index)
                                .extend([slot])
                        });
                        inputs.extend(array_indexed.indexes);
                    }
                    (
                        _,
                        InsnFieldType::SmallUInt(_)
                        | InsnFieldType::SmallSInt(_)
                        | InsnFieldType::InternedBigInt(_)
                        | InsnFieldType::U8(_)
                        | InsnFieldType::USize(_)
                        | InsnFieldType::Empty(_),
                    )
                    | (InsnFieldKind::Immediate | InsnFieldKind::BranchTarget, _) => {}
                }
            }
        }
        self.slots_assignments.assignments.push(Assignment {
            inputs,
            conditions,
            insns,
            source_location,
        });
    }
    fn simple_nary_big_expr<const N: usize>(
        &mut self,
        instantiated_module: InstantiatedModule,
        dest_ty: CanonicalType,
        inputs: [Expr<CanonicalType>; N],
        make_insns: impl FnOnce(
            StatePartIndex<StatePartKindBigSlots>,
            [StatePartIndex<StatePartKindBigSlots>; N],
        ) -> Vec<Insn>,
    ) -> CompiledValue<CanonicalType> {
        let inputs = inputs.map(|input| {
            let input = self.compile_expr(instantiated_module, input);
            let input = self
                .compiled_expr_to_value(input, instantiated_module.leaf_module().source_location());
            let TypeIndexRange {
                small_slots,
                big_slots,
            } = input.range;
            assert_eq!(small_slots.len.value, 0);
            assert_eq!(big_slots.len.value, 1);
            big_slots.start
        });
        let layout = CompiledTypeLayout::get(dest_ty);
        let range = self.insns.allocate_variable(&layout.layout);
        let TypeIndexRange {
            small_slots,
            big_slots,
        } = range;
        assert_eq!(small_slots.len.value, 0);
        assert_eq!(big_slots.len.value, 1);
        let retval = CompiledValue {
            layout,
            range,
            write: None,
        };
        self.add_assignment(
            Interned::default(),
            make_insns(big_slots.start, inputs),
            instantiated_module.leaf_module().source_location(),
        );
        retval
    }
    fn compile_expr(
        &mut self,
        instantiated_module: InstantiatedModule,
        expr: Expr<CanonicalType>,
    ) -> CompiledExpr<CanonicalType> {
        if let Some(&retval) = self.compiled_exprs.get(&expr) {
            return retval;
        }
        let mut cast_bit = |arg: Expr<CanonicalType>| {
            let src_signed = match Expr::ty(arg) {
                CanonicalType::UInt(_) => false,
                CanonicalType::SInt(_) => true,
                CanonicalType::Bool(_) => false,
                CanonicalType::Array(_) => unreachable!(),
                CanonicalType::Enum(_) => unreachable!(),
                CanonicalType::Bundle(_) => unreachable!(),
                CanonicalType::AsyncReset(_) => false,
                CanonicalType::SyncReset(_) => false,
                CanonicalType::Reset(_) => false,
                CanonicalType::Clock(_) => false,
            };
            let dest_signed = match Expr::ty(expr) {
                CanonicalType::UInt(_) => false,
                CanonicalType::SInt(_) => true,
                CanonicalType::Bool(_) => false,
                CanonicalType::Array(_) => unreachable!(),
                CanonicalType::Enum(_) => unreachable!(),
                CanonicalType::Bundle(_) => unreachable!(),
                CanonicalType::AsyncReset(_) => false,
                CanonicalType::SyncReset(_) => false,
                CanonicalType::Reset(_) => false,
                CanonicalType::Clock(_) => false,
            };
            self.simple_nary_big_expr(instantiated_module, Expr::ty(expr), [arg], |dest, [src]| {
                match (src_signed, dest_signed) {
                    (false, false) | (true, true) => {
                        vec![Insn::Copy { dest, src }]
                    }
                    (false, true) => vec![Insn::CastToSInt {
                        dest,
                        src,
                        dest_width: 1,
                    }],
                    (true, false) => vec![Insn::CastToUInt {
                        dest,
                        src,
                        dest_width: 1,
                    }],
                }
            })
            .into()
        };
        let retval: CompiledExpr<_> = match *Expr::expr_enum(expr) {
            ExprEnum::UIntLiteral(expr) => self
                .simple_nary_big_expr(
                    instantiated_module,
                    expr.ty().canonical(),
                    [],
                    |dest, []| {
                        vec![Insn::Const {
                            dest,
                            value: expr.to_bigint().intern_sized(),
                        }]
                    },
                )
                .into(),
            ExprEnum::SIntLiteral(expr) => self
                .simple_nary_big_expr(
                    instantiated_module,
                    expr.ty().canonical(),
                    [],
                    |dest, []| {
                        vec![Insn::Const {
                            dest,
                            value: expr.to_bigint().intern_sized(),
                        }]
                    },
                )
                .into(),
            ExprEnum::BoolLiteral(expr) => self
                .simple_nary_big_expr(instantiated_module, Bool.canonical(), [], |dest, []| {
                    vec![Insn::Const {
                        dest,
                        value: BigInt::from(expr).intern_sized(),
                    }]
                })
                .into(),
            ExprEnum::BundleLiteral(expr) => todo!(),
            ExprEnum::ArrayLiteral(expr) => todo!(),
            ExprEnum::EnumLiteral(expr) => todo!(),
            ExprEnum::Uninit(expr) => todo!(),
            ExprEnum::NotU(expr) => self
                .simple_nary_big_expr(
                    instantiated_module,
                    Expr::ty(expr.arg()).canonical(),
                    [Expr::canonical(expr.arg())],
                    |dest, [src]| {
                        vec![Insn::NotU {
                            dest,
                            src,
                            width: Expr::ty(expr.arg()).width(),
                        }]
                    },
                )
                .into(),
            ExprEnum::NotS(expr) => self
                .simple_nary_big_expr(
                    instantiated_module,
                    Expr::ty(expr.arg()).canonical(),
                    [Expr::canonical(expr.arg())],
                    |dest, [src]| vec![Insn::NotS { dest, src }],
                )
                .into(),
            ExprEnum::NotB(expr) => self
                .simple_nary_big_expr(
                    instantiated_module,
                    Expr::ty(expr.arg()).canonical(),
                    [Expr::canonical(expr.arg())],
                    |dest, [src]| {
                        vec![Insn::NotU {
                            dest,
                            src,
                            width: 1,
                        }]
                    },
                )
                .into(),
            ExprEnum::Neg(expr) => self
                .simple_nary_big_expr(
                    instantiated_module,
                    expr.ty().canonical(),
                    [Expr::canonical(expr.arg())],
                    |dest, [src]| vec![Insn::Neg { dest, src }],
                )
                .into(),
            ExprEnum::BitAndU(expr) => self
                .simple_nary_big_expr(
                    instantiated_module,
                    expr.ty().canonical(),
                    [Expr::canonical(expr.lhs()), Expr::canonical(expr.rhs())],
                    |dest, [lhs, rhs]| vec![Insn::And { dest, lhs, rhs }],
                )
                .into(),
            ExprEnum::BitAndS(expr) => self
                .simple_nary_big_expr(
                    instantiated_module,
                    expr.ty().canonical(),
                    [Expr::canonical(expr.lhs()), Expr::canonical(expr.rhs())],
                    |dest, [lhs, rhs]| vec![Insn::And { dest, lhs, rhs }],
                )
                .into(),
            ExprEnum::BitAndB(expr) => self
                .simple_nary_big_expr(
                    instantiated_module,
                    Bool.canonical(),
                    [Expr::canonical(expr.lhs()), Expr::canonical(expr.rhs())],
                    |dest, [lhs, rhs]| vec![Insn::And { dest, lhs, rhs }],
                )
                .into(),
            ExprEnum::BitOrU(expr) => self
                .simple_nary_big_expr(
                    instantiated_module,
                    expr.ty().canonical(),
                    [Expr::canonical(expr.lhs()), Expr::canonical(expr.rhs())],
                    |dest, [lhs, rhs]| vec![Insn::Or { dest, lhs, rhs }],
                )
                .into(),
            ExprEnum::BitOrS(expr) => self
                .simple_nary_big_expr(
                    instantiated_module,
                    expr.ty().canonical(),
                    [Expr::canonical(expr.lhs()), Expr::canonical(expr.rhs())],
                    |dest, [lhs, rhs]| vec![Insn::Or { dest, lhs, rhs }],
                )
                .into(),
            ExprEnum::BitOrB(expr) => self
                .simple_nary_big_expr(
                    instantiated_module,
                    Bool.canonical(),
                    [Expr::canonical(expr.lhs()), Expr::canonical(expr.rhs())],
                    |dest, [lhs, rhs]| vec![Insn::Or { dest, lhs, rhs }],
                )
                .into(),
            ExprEnum::BitXorU(expr) => self
                .simple_nary_big_expr(
                    instantiated_module,
                    expr.ty().canonical(),
                    [Expr::canonical(expr.lhs()), Expr::canonical(expr.rhs())],
                    |dest, [lhs, rhs]| vec![Insn::Xor { dest, lhs, rhs }],
                )
                .into(),
            ExprEnum::BitXorS(expr) => self
                .simple_nary_big_expr(
                    instantiated_module,
                    expr.ty().canonical(),
                    [Expr::canonical(expr.lhs()), Expr::canonical(expr.rhs())],
                    |dest, [lhs, rhs]| vec![Insn::Xor { dest, lhs, rhs }],
                )
                .into(),
            ExprEnum::BitXorB(expr) => self
                .simple_nary_big_expr(
                    instantiated_module,
                    Bool.canonical(),
                    [Expr::canonical(expr.lhs()), Expr::canonical(expr.rhs())],
                    |dest, [lhs, rhs]| vec![Insn::Xor { dest, lhs, rhs }],
                )
                .into(),
            ExprEnum::AddU(expr) => self
                .simple_nary_big_expr(
                    instantiated_module,
                    expr.ty().canonical(),
                    [Expr::canonical(expr.lhs()), Expr::canonical(expr.rhs())],
                    |dest, [lhs, rhs]| vec![Insn::Add { dest, lhs, rhs }],
                )
                .into(),
            ExprEnum::AddS(expr) => self
                .simple_nary_big_expr(
                    instantiated_module,
                    expr.ty().canonical(),
                    [Expr::canonical(expr.lhs()), Expr::canonical(expr.rhs())],
                    |dest, [lhs, rhs]| vec![Insn::Add { dest, lhs, rhs }],
                )
                .into(),
            ExprEnum::SubU(expr) => self
                .simple_nary_big_expr(
                    instantiated_module,
                    expr.ty().canonical(),
                    [Expr::canonical(expr.lhs()), Expr::canonical(expr.rhs())],
                    |dest, [lhs, rhs]| {
                        vec![Insn::SubU {
                            dest,
                            lhs,
                            rhs,
                            dest_width: expr.ty().width(),
                        }]
                    },
                )
                .into(),
            ExprEnum::SubS(expr) => self
                .simple_nary_big_expr(
                    instantiated_module,
                    expr.ty().canonical(),
                    [Expr::canonical(expr.lhs()), Expr::canonical(expr.rhs())],
                    |dest, [lhs, rhs]| vec![Insn::SubS { dest, lhs, rhs }],
                )
                .into(),
            ExprEnum::MulU(expr) => self
                .simple_nary_big_expr(
                    instantiated_module,
                    expr.ty().canonical(),
                    [Expr::canonical(expr.lhs()), Expr::canonical(expr.rhs())],
                    |dest, [lhs, rhs]| vec![Insn::Mul { dest, lhs, rhs }],
                )
                .into(),
            ExprEnum::MulS(expr) => self
                .simple_nary_big_expr(
                    instantiated_module,
                    expr.ty().canonical(),
                    [Expr::canonical(expr.lhs()), Expr::canonical(expr.rhs())],
                    |dest, [lhs, rhs]| vec![Insn::Mul { dest, lhs, rhs }],
                )
                .into(),
            ExprEnum::DivU(expr) => self
                .simple_nary_big_expr(
                    instantiated_module,
                    expr.ty().canonical(),
                    [Expr::canonical(expr.lhs()), Expr::canonical(expr.rhs())],
                    |dest, [lhs, rhs]| vec![Insn::Div { dest, lhs, rhs }],
                )
                .into(),
            ExprEnum::DivS(expr) => self
                .simple_nary_big_expr(
                    instantiated_module,
                    expr.ty().canonical(),
                    [Expr::canonical(expr.lhs()), Expr::canonical(expr.rhs())],
                    |dest, [lhs, rhs]| vec![Insn::Div { dest, lhs, rhs }],
                )
                .into(),
            ExprEnum::RemU(expr) => self
                .simple_nary_big_expr(
                    instantiated_module,
                    expr.ty().canonical(),
                    [Expr::canonical(expr.lhs()), Expr::canonical(expr.rhs())],
                    |dest, [lhs, rhs]| vec![Insn::Rem { dest, lhs, rhs }],
                )
                .into(),
            ExprEnum::RemS(expr) => self
                .simple_nary_big_expr(
                    instantiated_module,
                    expr.ty().canonical(),
                    [Expr::canonical(expr.lhs()), Expr::canonical(expr.rhs())],
                    |dest, [lhs, rhs]| vec![Insn::Rem { dest, lhs, rhs }],
                )
                .into(),
            ExprEnum::DynShlU(expr) => self
                .simple_nary_big_expr(
                    instantiated_module,
                    expr.ty().canonical(),
                    [Expr::canonical(expr.lhs()), Expr::canonical(expr.rhs())],
                    |dest, [lhs, rhs]| vec![Insn::DynShl { dest, lhs, rhs }],
                )
                .into(),
            ExprEnum::DynShlS(expr) => self
                .simple_nary_big_expr(
                    instantiated_module,
                    expr.ty().canonical(),
                    [Expr::canonical(expr.lhs()), Expr::canonical(expr.rhs())],
                    |dest, [lhs, rhs]| vec![Insn::DynShl { dest, lhs, rhs }],
                )
                .into(),
            ExprEnum::DynShrU(expr) => self
                .simple_nary_big_expr(
                    instantiated_module,
                    expr.ty().canonical(),
                    [Expr::canonical(expr.lhs()), Expr::canonical(expr.rhs())],
                    |dest, [lhs, rhs]| vec![Insn::DynShr { dest, lhs, rhs }],
                )
                .into(),
            ExprEnum::DynShrS(expr) => self
                .simple_nary_big_expr(
                    instantiated_module,
                    expr.ty().canonical(),
                    [Expr::canonical(expr.lhs()), Expr::canonical(expr.rhs())],
                    |dest, [lhs, rhs]| vec![Insn::DynShr { dest, lhs, rhs }],
                )
                .into(),
            ExprEnum::FixedShlU(expr) => self
                .simple_nary_big_expr(
                    instantiated_module,
                    expr.ty().canonical(),
                    [Expr::canonical(expr.lhs())],
                    |dest, [lhs]| {
                        vec![Insn::Shl {
                            dest,
                            lhs,
                            rhs: expr.rhs(),
                        }]
                    },
                )
                .into(),
            ExprEnum::FixedShlS(expr) => self
                .simple_nary_big_expr(
                    instantiated_module,
                    expr.ty().canonical(),
                    [Expr::canonical(expr.lhs())],
                    |dest, [lhs]| {
                        vec![Insn::Shl {
                            dest,
                            lhs,
                            rhs: expr.rhs(),
                        }]
                    },
                )
                .into(),
            ExprEnum::FixedShrU(expr) => self
                .simple_nary_big_expr(
                    instantiated_module,
                    expr.ty().canonical(),
                    [Expr::canonical(expr.lhs())],
                    |dest, [lhs]| {
                        vec![Insn::Shr {
                            dest,
                            lhs,
                            rhs: expr.rhs(),
                        }]
                    },
                )
                .into(),
            ExprEnum::FixedShrS(expr) => self
                .simple_nary_big_expr(
                    instantiated_module,
                    expr.ty().canonical(),
                    [Expr::canonical(expr.lhs())],
                    |dest, [lhs]| {
                        vec![Insn::Shr {
                            dest,
                            lhs,
                            rhs: expr.rhs(),
                        }]
                    },
                )
                .into(),
            ExprEnum::CmpLtB(expr) => self
                .simple_nary_big_expr(
                    instantiated_module,
                    Bool.canonical(),
                    [Expr::canonical(expr.lhs()), Expr::canonical(expr.rhs())],
                    |dest, [lhs, rhs]| vec![Insn::CmpLt { dest, lhs, rhs }],
                )
                .into(),
            ExprEnum::CmpLeB(expr) => self
                .simple_nary_big_expr(
                    instantiated_module,
                    Bool.canonical(),
                    [Expr::canonical(expr.lhs()), Expr::canonical(expr.rhs())],
                    |dest, [lhs, rhs]| vec![Insn::CmpLe { dest, lhs, rhs }],
                )
                .into(),
            ExprEnum::CmpGtB(expr) => self
                .simple_nary_big_expr(
                    instantiated_module,
                    Bool.canonical(),
                    // swap both comparison direction and lhs/rhs
                    [Expr::canonical(expr.rhs()), Expr::canonical(expr.lhs())],
                    |dest, [lhs, rhs]| vec![Insn::CmpLt { dest, lhs, rhs }],
                )
                .into(),
            ExprEnum::CmpGeB(expr) => self
                .simple_nary_big_expr(
                    instantiated_module,
                    Bool.canonical(),
                    // swap both comparison direction and lhs/rhs
                    [Expr::canonical(expr.rhs()), Expr::canonical(expr.lhs())],
                    |dest, [lhs, rhs]| vec![Insn::CmpLe { dest, lhs, rhs }],
                )
                .into(),
            ExprEnum::CmpEqB(expr) => self
                .simple_nary_big_expr(
                    instantiated_module,
                    Bool.canonical(),
                    [Expr::canonical(expr.lhs()), Expr::canonical(expr.rhs())],
                    |dest, [lhs, rhs]| vec![Insn::CmpEq { dest, lhs, rhs }],
                )
                .into(),
            ExprEnum::CmpNeB(expr) => self
                .simple_nary_big_expr(
                    instantiated_module,
                    Bool.canonical(),
                    [Expr::canonical(expr.lhs()), Expr::canonical(expr.rhs())],
                    |dest, [lhs, rhs]| vec![Insn::CmpNe { dest, lhs, rhs }],
                )
                .into(),
            ExprEnum::CmpLtU(expr) => self
                .simple_nary_big_expr(
                    instantiated_module,
                    Bool.canonical(),
                    [Expr::canonical(expr.lhs()), Expr::canonical(expr.rhs())],
                    |dest, [lhs, rhs]| vec![Insn::CmpLt { dest, lhs, rhs }],
                )
                .into(),
            ExprEnum::CmpLeU(expr) => self
                .simple_nary_big_expr(
                    instantiated_module,
                    Bool.canonical(),
                    [Expr::canonical(expr.lhs()), Expr::canonical(expr.rhs())],
                    |dest, [lhs, rhs]| vec![Insn::CmpLe { dest, lhs, rhs }],
                )
                .into(),
            ExprEnum::CmpGtU(expr) => self
                .simple_nary_big_expr(
                    instantiated_module,
                    Bool.canonical(),
                    // swap both comparison direction and lhs/rhs
                    [Expr::canonical(expr.rhs()), Expr::canonical(expr.lhs())],
                    |dest, [lhs, rhs]| vec![Insn::CmpLt { dest, lhs, rhs }],
                )
                .into(),
            ExprEnum::CmpGeU(expr) => self
                .simple_nary_big_expr(
                    instantiated_module,
                    Bool.canonical(),
                    // swap both comparison direction and lhs/rhs
                    [Expr::canonical(expr.rhs()), Expr::canonical(expr.lhs())],
                    |dest, [lhs, rhs]| vec![Insn::CmpLe { dest, lhs, rhs }],
                )
                .into(),
            ExprEnum::CmpEqU(expr) => self
                .simple_nary_big_expr(
                    instantiated_module,
                    Bool.canonical(),
                    [Expr::canonical(expr.lhs()), Expr::canonical(expr.rhs())],
                    |dest, [lhs, rhs]| vec![Insn::CmpEq { dest, lhs, rhs }],
                )
                .into(),
            ExprEnum::CmpNeU(expr) => self
                .simple_nary_big_expr(
                    instantiated_module,
                    Bool.canonical(),
                    [Expr::canonical(expr.lhs()), Expr::canonical(expr.rhs())],
                    |dest, [lhs, rhs]| vec![Insn::CmpNe { dest, lhs, rhs }],
                )
                .into(),
            ExprEnum::CmpLtS(expr) => self
                .simple_nary_big_expr(
                    instantiated_module,
                    Bool.canonical(),
                    [Expr::canonical(expr.lhs()), Expr::canonical(expr.rhs())],
                    |dest, [lhs, rhs]| vec![Insn::CmpLt { dest, lhs, rhs }],
                )
                .into(),
            ExprEnum::CmpLeS(expr) => self
                .simple_nary_big_expr(
                    instantiated_module,
                    Bool.canonical(),
                    [Expr::canonical(expr.lhs()), Expr::canonical(expr.rhs())],
                    |dest, [lhs, rhs]| vec![Insn::CmpLe { dest, lhs, rhs }],
                )
                .into(),
            ExprEnum::CmpGtS(expr) => self
                .simple_nary_big_expr(
                    instantiated_module,
                    Bool.canonical(),
                    // swap both comparison direction and lhs/rhs
                    [Expr::canonical(expr.rhs()), Expr::canonical(expr.lhs())],
                    |dest, [lhs, rhs]| vec![Insn::CmpLt { dest, lhs, rhs }],
                )
                .into(),
            ExprEnum::CmpGeS(expr) => self
                .simple_nary_big_expr(
                    instantiated_module,
                    Bool.canonical(),
                    // swap both comparison direction and lhs/rhs
                    [Expr::canonical(expr.rhs()), Expr::canonical(expr.lhs())],
                    |dest, [lhs, rhs]| vec![Insn::CmpLe { dest, lhs, rhs }],
                )
                .into(),
            ExprEnum::CmpEqS(expr) => self
                .simple_nary_big_expr(
                    instantiated_module,
                    Bool.canonical(),
                    [Expr::canonical(expr.lhs()), Expr::canonical(expr.rhs())],
                    |dest, [lhs, rhs]| vec![Insn::CmpEq { dest, lhs, rhs }],
                )
                .into(),
            ExprEnum::CmpNeS(expr) => self
                .simple_nary_big_expr(
                    instantiated_module,
                    Bool.canonical(),
                    [Expr::canonical(expr.lhs()), Expr::canonical(expr.rhs())],
                    |dest, [lhs, rhs]| vec![Insn::CmpNe { dest, lhs, rhs }],
                )
                .into(),
            ExprEnum::CastUIntToUInt(expr) => self
                .simple_nary_big_expr(
                    instantiated_module,
                    expr.ty().canonical(),
                    [Expr::canonical(expr.arg())],
                    |dest, [src]| {
                        vec![Insn::CastToUInt {
                            dest,
                            src,
                            dest_width: expr.ty().width(),
                        }]
                    },
                )
                .into(),
            ExprEnum::CastUIntToSInt(expr) => self
                .simple_nary_big_expr(
                    instantiated_module,
                    expr.ty().canonical(),
                    [Expr::canonical(expr.arg())],
                    |dest, [src]| {
                        vec![Insn::CastToSInt {
                            dest,
                            src,
                            dest_width: expr.ty().width(),
                        }]
                    },
                )
                .into(),
            ExprEnum::CastSIntToUInt(expr) => self
                .simple_nary_big_expr(
                    instantiated_module,
                    expr.ty().canonical(),
                    [Expr::canonical(expr.arg())],
                    |dest, [src]| {
                        vec![Insn::CastToUInt {
                            dest,
                            src,
                            dest_width: expr.ty().width(),
                        }]
                    },
                )
                .into(),
            ExprEnum::CastSIntToSInt(expr) => self
                .simple_nary_big_expr(
                    instantiated_module,
                    expr.ty().canonical(),
                    [Expr::canonical(expr.arg())],
                    |dest, [src]| {
                        vec![Insn::CastToSInt {
                            dest,
                            src,
                            dest_width: expr.ty().width(),
                        }]
                    },
                )
                .into(),
            ExprEnum::CastBoolToUInt(expr) => cast_bit(Expr::canonical(expr.arg())),
            ExprEnum::CastBoolToSInt(expr) => cast_bit(Expr::canonical(expr.arg())),
            ExprEnum::CastUIntToBool(expr) => cast_bit(Expr::canonical(expr.arg())),
            ExprEnum::CastSIntToBool(expr) => cast_bit(Expr::canonical(expr.arg())),
            ExprEnum::CastBoolToSyncReset(expr) => cast_bit(Expr::canonical(expr.arg())),
            ExprEnum::CastUIntToSyncReset(expr) => cast_bit(Expr::canonical(expr.arg())),
            ExprEnum::CastSIntToSyncReset(expr) => cast_bit(Expr::canonical(expr.arg())),
            ExprEnum::CastBoolToAsyncReset(expr) => cast_bit(Expr::canonical(expr.arg())),
            ExprEnum::CastUIntToAsyncReset(expr) => cast_bit(Expr::canonical(expr.arg())),
            ExprEnum::CastSIntToAsyncReset(expr) => cast_bit(Expr::canonical(expr.arg())),
            ExprEnum::CastSyncResetToBool(expr) => cast_bit(Expr::canonical(expr.arg())),
            ExprEnum::CastSyncResetToUInt(expr) => cast_bit(Expr::canonical(expr.arg())),
            ExprEnum::CastSyncResetToSInt(expr) => cast_bit(Expr::canonical(expr.arg())),
            ExprEnum::CastSyncResetToReset(expr) => cast_bit(Expr::canonical(expr.arg())),
            ExprEnum::CastAsyncResetToBool(expr) => cast_bit(Expr::canonical(expr.arg())),
            ExprEnum::CastAsyncResetToUInt(expr) => cast_bit(Expr::canonical(expr.arg())),
            ExprEnum::CastAsyncResetToSInt(expr) => cast_bit(Expr::canonical(expr.arg())),
            ExprEnum::CastAsyncResetToReset(expr) => cast_bit(Expr::canonical(expr.arg())),
            ExprEnum::CastResetToBool(expr) => cast_bit(Expr::canonical(expr.arg())),
            ExprEnum::CastResetToUInt(expr) => cast_bit(Expr::canonical(expr.arg())),
            ExprEnum::CastResetToSInt(expr) => cast_bit(Expr::canonical(expr.arg())),
            ExprEnum::CastBoolToClock(expr) => cast_bit(Expr::canonical(expr.arg())),
            ExprEnum::CastUIntToClock(expr) => cast_bit(Expr::canonical(expr.arg())),
            ExprEnum::CastSIntToClock(expr) => cast_bit(Expr::canonical(expr.arg())),
            ExprEnum::CastClockToBool(expr) => cast_bit(Expr::canonical(expr.arg())),
            ExprEnum::CastClockToUInt(expr) => cast_bit(Expr::canonical(expr.arg())),
            ExprEnum::CastClockToSInt(expr) => cast_bit(Expr::canonical(expr.arg())),
            ExprEnum::FieldAccess(expr) => self
                .compile_expr(instantiated_module, Expr::canonical(expr.base()))
                .map_ty(Bundle::from_canonical)
                .field_by_index(expr.field_index()),
            ExprEnum::VariantAccess(expr) => todo!(),
            ExprEnum::ArrayIndex(expr) => self
                .compile_expr(instantiated_module, Expr::canonical(expr.base()))
                .map_ty(Array::from_canonical)
                .element(expr.element_index()),
            ExprEnum::DynArrayIndex(expr) => todo!(),
            ExprEnum::ReduceBitAndU(expr) => if Expr::ty(expr.arg()).width() == 0 {
                self.compile_expr(instantiated_module, Expr::canonical(true.to_expr()))
            } else {
                self.compile_expr(
                    instantiated_module,
                    Expr::canonical(
                        expr.arg()
                            .cmp_eq(Expr::ty(expr.arg()).from_int_wrapping(-1)),
                    ),
                )
            }
            .into(),
            ExprEnum::ReduceBitAndS(expr) => if Expr::ty(expr.arg()).width() == 0 {
                self.compile_expr(instantiated_module, Expr::canonical(true.to_expr()))
            } else {
                self.compile_expr(
                    instantiated_module,
                    Expr::canonical(
                        expr.arg()
                            .cmp_eq(Expr::ty(expr.arg()).from_int_wrapping(-1)),
                    ),
                )
            }
            .into(),
            ExprEnum::ReduceBitOrU(expr) => if Expr::ty(expr.arg()).width() == 0 {
                self.compile_expr(instantiated_module, Expr::canonical(false.to_expr()))
            } else {
                self.compile_expr(
                    instantiated_module,
                    Expr::canonical(expr.arg().cmp_ne(Expr::ty(expr.arg()).from_int_wrapping(0))),
                )
            }
            .into(),
            ExprEnum::ReduceBitOrS(expr) => if Expr::ty(expr.arg()).width() == 0 {
                self.compile_expr(instantiated_module, Expr::canonical(false.to_expr()))
            } else {
                self.compile_expr(
                    instantiated_module,
                    Expr::canonical(expr.arg().cmp_ne(Expr::ty(expr.arg()).from_int_wrapping(0))),
                )
            }
            .into(),
            ExprEnum::ReduceBitXorU(expr) => self
                .simple_nary_big_expr(
                    instantiated_module,
                    UInt::<1>::TYPE.canonical(),
                    [Expr::canonical(expr.arg())],
                    |dest, [src]| {
                        vec![Insn::ReduceBitXor {
                            dest,
                            src,
                            input_width: Expr::ty(expr.arg()).width(),
                        }]
                    },
                )
                .into(),
            ExprEnum::ReduceBitXorS(expr) => self
                .simple_nary_big_expr(
                    instantiated_module,
                    UInt::<1>::TYPE.canonical(),
                    [Expr::canonical(expr.arg())],
                    |dest, [src]| {
                        vec![Insn::ReduceBitXor {
                            dest,
                            src,
                            input_width: Expr::ty(expr.arg()).width(),
                        }]
                    },
                )
                .into(),
            ExprEnum::SliceUInt(expr) => todo!(),
            ExprEnum::SliceSInt(expr) => todo!(),
            ExprEnum::CastToBits(expr) => todo!(),
            ExprEnum::CastBitsTo(expr) => todo!(),
            ExprEnum::ModuleIO(expr) => self
                .compile_value(TargetInInstantiatedModule {
                    instantiated_module,
                    target: expr.into(),
                })
                .into(),
            ExprEnum::Instance(expr) => self
                .compile_value(TargetInInstantiatedModule {
                    instantiated_module,
                    target: expr.into(),
                })
                .into(),
            ExprEnum::Wire(expr) => self
                .compile_value(TargetInInstantiatedModule {
                    instantiated_module,
                    target: expr.into(),
                })
                .into(),
            ExprEnum::Reg(expr) => self
                .compile_value(TargetInInstantiatedModule {
                    instantiated_module,
                    target: expr.into(),
                })
                .into(),
            ExprEnum::MemPort(expr) => self
                .compile_value(TargetInInstantiatedModule {
                    instantiated_module,
                    target: expr.into(),
                })
                .into(),
        };
        self.compiled_exprs.insert(expr, retval);
        retval
    }
    fn compile_simple_connect(
        &mut self,
        conditions: Interned<[Cond]>,
        lhs: CompiledExpr<CanonicalType>,
        rhs: CompiledValue<CanonicalType>,
        source_location: SourceLocation,
    ) {
        let CompiledExpr {
            static_part: lhs_static_part,
            indexes,
        } = lhs;
        let (lhs_layout, lhs_range) = lhs_static_part.write();
        assert!(
            lhs_layout.ty.is_passive(),
            "invalid expression passed to compile_simple_connect -- type must be passive",
        );
        let TypeIndexRange {
            small_slots,
            big_slots,
        } = lhs_range;
        self.add_assignment(
            conditions,
            small_slots
                .iter()
                .zip(rhs.range.small_slots.iter())
                .map(|(base, src)| {
                    if indexes.small_slots.is_empty() {
                        Insn::CopySmall { dest: base, src }
                    } else {
                        Insn::WriteSmallIndexed {
                            dest: StatePartArrayIndexed {
                                base,
                                indexes: indexes.small_slots,
                            },
                            src,
                        }
                    }
                })
                .chain(
                    big_slots
                        .iter()
                        .zip(rhs.range.big_slots.iter())
                        .map(|(base, src)| {
                            if indexes.big_slots.is_empty() {
                                Insn::Copy { dest: base, src }
                            } else {
                                Insn::WriteIndexed {
                                    dest: StatePartArrayIndexed {
                                        base,
                                        indexes: indexes.big_slots,
                                    },
                                    src,
                                }
                            }
                        }),
                ),
            source_location,
        );
    }
    fn compile_connect(
        &mut self,
        lhs_instantiated_module: InstantiatedModule,
        lhs_conditions: Interned<[Cond]>,
        lhs: Expr<CanonicalType>,
        rhs_instantiated_module: InstantiatedModule,
        rhs_conditions: Interned<[Cond]>,
        mut rhs: Expr<CanonicalType>,
        source_location: SourceLocation,
    ) {
        if Expr::ty(lhs) != Expr::ty(rhs) || !Expr::ty(lhs).is_passive() {
            match Expr::ty(lhs) {
                CanonicalType::UInt(lhs_ty) => {
                    rhs = Expr::canonical(Expr::<UInt>::from_canonical(rhs).cast_to(lhs_ty));
                }
                CanonicalType::SInt(lhs_ty) => {
                    rhs = Expr::canonical(Expr::<SInt>::from_canonical(rhs).cast_to(lhs_ty));
                }
                CanonicalType::Bool(_) => unreachable!(),
                CanonicalType::Array(lhs_ty) => {
                    let CanonicalType::Array(rhs_ty) = Expr::ty(rhs) else {
                        unreachable!();
                    };
                    assert_eq!(lhs_ty.len(), rhs_ty.len());
                    let lhs = Expr::<Array>::from_canonical(lhs);
                    let rhs = Expr::<Array>::from_canonical(rhs);
                    for index in 0..lhs_ty.len() {
                        self.compile_connect(
                            lhs_instantiated_module,
                            lhs_conditions,
                            lhs[index],
                            rhs_instantiated_module,
                            rhs_conditions,
                            rhs[index],
                            source_location,
                        );
                    }
                    return;
                }
                CanonicalType::Enum(lhs_ty) => {
                    let CanonicalType::Enum(rhs_ty) = Expr::ty(rhs) else {
                        unreachable!();
                    };
                    todo!("handle connect with different enum types");
                }
                CanonicalType::Bundle(lhs_ty) => {
                    let CanonicalType::Bundle(rhs_ty) = Expr::ty(rhs) else {
                        unreachable!();
                    };
                    assert_eq!(lhs_ty.fields().len(), rhs_ty.fields().len());
                    let lhs = Expr::<Bundle>::from_canonical(lhs);
                    let rhs = Expr::<Bundle>::from_canonical(rhs);
                    for (
                        field_index,
                        (
                            BundleField {
                                name,
                                flipped,
                                ty: _,
                            },
                            rhs_field,
                        ),
                    ) in lhs_ty.fields().into_iter().zip(rhs_ty.fields()).enumerate()
                    {
                        assert_eq!(name, rhs_field.name);
                        assert_eq!(flipped, rhs_field.flipped);
                        let lhs_expr = ops::FieldAccess::new_by_index(lhs, field_index).to_expr();
                        let rhs_expr = ops::FieldAccess::new_by_index(rhs, field_index).to_expr();
                        if flipped {
                            // swap lhs/rhs
                            self.compile_connect(
                                rhs_instantiated_module,
                                rhs_conditions,
                                rhs_expr,
                                lhs_instantiated_module,
                                lhs_conditions,
                                lhs_expr,
                                source_location,
                            );
                        } else {
                            self.compile_connect(
                                lhs_instantiated_module,
                                lhs_conditions,
                                lhs_expr,
                                rhs_instantiated_module,
                                rhs_conditions,
                                rhs_expr,
                                source_location,
                            );
                        }
                    }
                    return;
                }
                CanonicalType::AsyncReset(_) => unreachable!(),
                CanonicalType::SyncReset(_) => unreachable!(),
                CanonicalType::Reset(_) => unreachable!(),
                CanonicalType::Clock(_) => unreachable!(),
            }
        }
        let Some(target) = lhs.target() else {
            unreachable!("connect lhs must have target");
        };
        let lhs_decl_conditions = self.decl_conditions[&TargetInInstantiatedModule {
            instantiated_module: lhs_instantiated_module,
            target: target.base().into(),
        }];
        let lhs = self.compile_expr(lhs_instantiated_module, lhs);
        let rhs = self.compile_expr(rhs_instantiated_module, rhs);
        let rhs = self.compiled_expr_to_value(rhs, source_location);
        self.compile_simple_connect(
            lhs_conditions[lhs_decl_conditions.len()..].intern(),
            lhs,
            rhs,
            source_location,
        );
    }
    fn compile_declaration(
        &mut self,
        declaration: StmtDeclaration,
        parent_module: Interned<InstantiatedModule>,
        conditions: Interned<[Cond]>,
    ) {
        let target_base: TargetBase = match &declaration {
            StmtDeclaration::Wire(v) => v.wire.into(),
            StmtDeclaration::Reg(v) => v.reg.into(),
            StmtDeclaration::Instance(v) => v.instance.into(),
        };
        let target = TargetInInstantiatedModule {
            instantiated_module: *parent_module,
            target: target_base.into(),
        };
        self.decl_conditions.insert(target, conditions);
        self.compile_value(target);
        match declaration {
            StmtDeclaration::Wire(StmtWire { annotations, wire }) => {}
            StmtDeclaration::Reg(StmtReg { annotations, reg }) => {
                todo!();
            }
            StmtDeclaration::Instance(StmtInstance {
                annotations,
                instance,
            }) => {
                let inner_instantiated_module = InstantiatedModule::Child {
                    parent: parent_module,
                    instance: instance.intern_sized(),
                }
                .intern_sized();
                let instance_expr = instance.to_expr();
                self.compile_module(inner_instantiated_module);
                for (field_index, module_io) in
                    instance.instantiated().module_io().into_iter().enumerate()
                {
                    let instance_field =
                        ops::FieldAccess::new_by_index(instance_expr, field_index).to_expr();
                    match Expr::flow(instance_field) {
                        Flow::Source => {
                            // we need to supply the value to the instance since the
                            // parent module expects to read from the instance
                            self.compile_connect(
                                *parent_module,
                                conditions,
                                instance_field,
                                *inner_instantiated_module,
                                Interned::default(),
                                module_io.module_io.to_expr(),
                                instance.source_location(),
                            );
                        }
                        Flow::Sink => {
                            // we need to take the value from the instance since the
                            // parent module expects to write to the instance
                            self.compile_connect(
                                *inner_instantiated_module,
                                Interned::default(),
                                module_io.module_io.to_expr(),
                                *parent_module,
                                conditions,
                                instance_field,
                                instance.source_location(),
                            );
                        }
                        Flow::Duplex => unreachable!(),
                    }
                }
            }
        }
    }
    fn compile_block(
        &mut self,
        parent_module: Interned<InstantiatedModule>,
        block: Block,
        conditions: Interned<[Cond]>,
    ) {
        let Block { memories, stmts } = block;
        for memory in memories {
            todo!("implement memory");
        }
        for stmt in stmts {
            match stmt {
                Stmt::Connect(StmtConnect {
                    lhs,
                    rhs,
                    source_location,
                }) => self.compile_connect(
                    *parent_module,
                    conditions,
                    lhs,
                    *parent_module,
                    conditions,
                    rhs,
                    source_location,
                ),
                Stmt::Formal(StmtFormal { .. }) => todo!("implement simulating formal statements"),
                Stmt::If(StmtIf {
                    cond,
                    source_location,
                    blocks: [then_block, else_block],
                }) => {
                    let cond = self
                        .compile_expr(*parent_module, Expr::canonical(cond))
                        .map_ty(Bool::from_canonical);
                    self.compile_block(
                        parent_module,
                        then_block,
                        Interned::from_iter(conditions.iter().copied().chain([Cond {
                            body: CondBody::IfTrue { cond },
                            source_location,
                        }])),
                    );
                    self.compile_block(
                        parent_module,
                        else_block,
                        Interned::from_iter(conditions.iter().copied().chain([Cond {
                            body: CondBody::IfFalse { cond },
                            source_location,
                        }])),
                    );
                }
                Stmt::Match(StmtMatch {
                    expr,
                    source_location,
                    blocks,
                }) => {
                    let enum_expr = self
                        .compile_expr(*parent_module, Expr::canonical(expr))
                        .map_ty(Enum::from_canonical);
                    for (variant_index, block) in blocks.into_iter().enumerate() {
                        self.compile_block(
                            parent_module,
                            block,
                            Interned::from_iter(conditions.iter().copied().chain([Cond {
                                body: CondBody::MatchArm {
                                    enum_expr,
                                    variant_index,
                                },
                                source_location,
                            }])),
                        );
                    }
                }
                Stmt::Declaration(declaration) => {
                    self.compile_declaration(declaration, parent_module, conditions);
                }
            }
        }
    }
    fn compile_module(&mut self, module: Interned<InstantiatedModule>) -> &CompiledModule {
        let module_io = module
            .leaf_module()
            .module_io()
            .iter()
            .map(
                |&AnnotatedModuleIO {
                     annotations: _,
                     module_io,
                 }| {
                    self.compile_value(TargetInInstantiatedModule {
                        instantiated_module: *module,
                        target: Target::from(module_io),
                    })
                },
            )
            .collect();
        match module.leaf_module().body() {
            ModuleBody::Normal(NormalModuleBody { body }) => {
                self.compile_block(module, body, Interned::default());
            }
            ModuleBody::Extern(_extern_module_body) => {
                todo!("simulating extern module: {:?}", module);
            }
        }
        let hashbrown::hash_map::Entry::Vacant(entry) = self.modules.entry(*module) else {
            unreachable!("compiled same instantiated module twice");
        };
        entry.insert(CompiledModule { module_io })
    }
    pub fn compile(mut self) -> Compiled {
        self.compile_module(InstantiatedModule::Base(self.base_module).intern_sized());
        todo!("handle self.slots_assignments");
        Compiled {
            insns: Insns::from(self.insns).intern_sized(),
            modules: self.modules,
        }
    }
}

#[derive(Debug)]
struct CompiledModule {
    module_io: Interned<[CompiledValue<CanonicalType>]>,
}

#[derive(Debug)]
pub struct Compiled {
    insns: Interned<Insns<InsnsBuildingDone>>,
    modules: HashMap<InstantiatedModule, CompiledModule>,
}

impl Compiled {
    pub fn new<T: BundleType>(module: Module<T>) -> Self {
        Compiler::new(module.canonical().intern()).compile()
    }
}

#[derive(Debug)]
pub struct Simulation {
    state: interpreter::State,
}
